@@ -38,16 +38,18 @@ module Database.Esqueleto
     -- $gettingstarted
 
     -- * @esqueleto@'s Language
-    Esqueleto( where_, on, groupBy, orderBy, rand, asc, desc, limit, offset, having
+    Esqueleto( where_, on, groupBy, orderBy, rand, asc, desc, limit, offset
+             , distinct, distinctOn, don, distinctOnOrderBy, having, locking
              , sub_select, sub_selectDistinct, (^.), (?.)
-             , val, isNothing, just, nothing, joinV, countRows, count, not_
-             , (==.), (>=.), (>.), (<=.), (<.), (!=.), (&&.), (||.)
+             , val, isNothing, just, nothing, joinV
+             , countRows, count, countDistinct
+             , not_, (==.), (>=.), (>.), (<=.), (<.), (!=.), (&&.), (||.)
              , (+.), (-.), (/.), (*.)
              , random_, round_, ceiling_, floor_
-             , min_, max_, sum_, avg_, lower_
+             , min_, max_, sum_, avg_, castNum, castNumM
              , coalesce, coalesceDefault
-             , like, (%), concat_, (++.)
-             , subList_select, subList_selectDistinct, valList
+             , lower_, like, ilike, (%), concat_, (++.), castString
+             , subList_select, subList_selectDistinct, valList, justList
              , in_, notIn, exists, notExists
              , set, (=.), (+=.), (-=.), (*=.), (/=.)
              , case_ )
@@ -59,6 +61,9 @@ module Database.Esqueleto
   , unValue
   , ValueList(..)
   , OrderBy
+  , DistinctOn
+  , LockingKind(..)
+  , SqlString
     -- ** Joins
   , InnerJoin(..)
   , CrossJoin(..)
@@ -84,6 +89,9 @@ module Database.Esqueleto
   , (<#)
   , (<&>)
 
+    -- * RDBMS-specific modules
+    -- $rdbmsSpecificModules
+
     -- * Helpers
   , valkey
   , valJ
@@ -129,7 +137,7 @@ import qualified Database.Persist
 --   * Support the mostly used SQL features.  We'd like you to be
 --   able to use @esqueleto@ for all of your queries, no
 --   exceptions.  Send a pull request or open an issue on our
---   project page (<https://github.com/meteficha/esqueleto>) if
+--   project page (<https://github.com/prowdsponsor/esqueleto>) if
 --   there's anything missing that you'd like to see.
 --
 --   * Be as type-safe as possible.  We strive to provide as many
@@ -214,8 +222,8 @@ import qualified Database.Persist
 -- In @esqueleto@, we may write the same query above as:
 --
 -- @
--- select $
--- from $ \\p -> do
+-- 'select' $
+-- 'from' $ \\p -> do
 -- 'where_' (p '^.' PersonName '==.' 'val' \"John\")
 -- return p
 -- @
@@ -238,14 +246,14 @@ import qualified Database.Persist
 -- In @esqueleto@, we may write the same query above as:
 --
 -- @
--- select $
--- from $ \\p -> do
--- where_ (p ^. PersonAge '>=.' 'just' (val 18))
+-- 'select' $
+-- 'from' $ \\p -> do
+-- 'where_' (p '^.' PersonAge '>=.' 'just' ('val' 18))
 -- return p
 -- @
 --
 -- Since @age@ is an optional @Person@ field, we use 'just' lift
--- @val 18 :: SqlExpr (Value Int)@ into @just (val 18) ::
+-- @'val' 18 :: SqlExpr (Value Int)@ into @just ('val' 18) ::
 -- SqlExpr (Value (Maybe Int))@.
 --
 -- Implicit joins are represented by tuples.  For example, to get
@@ -261,10 +269,10 @@ import qualified Database.Persist
 -- In @esqueleto@, we may write the same query above as:
 --
 -- @
--- select $
--- from $ \\(b, p) -> do
--- where_ (b ^. BlogPostAuthorId ==. p ^. PersonId)
--- 'orderBy' ['asc' (b ^. BlogPostTitle)]
+-- 'select' $
+-- 'from' $ \\(b, p) -> do
+-- 'where_' (b '^.' BlogPostAuthorId '==.' p '^.' PersonId)
+-- 'orderBy' ['asc' (b '^.' BlogPostTitle)]
 -- return (b, p)
 -- @
 --
@@ -281,10 +289,10 @@ import qualified Database.Persist
 -- In @esqueleto@, we may write the same query above as:
 --
 -- @
--- select $
--- from $ \\(p ``LeftOuterJoin`` mb) -> do
--- 'on' (just (p ^. PersonId) ==. mb '?.' BlogPostAuthorId)
--- orderBy [asc (p ^. PersonName), asc (mb '?.' BlogPostTitle)]
+-- 'select' $
+-- 'from' $ \\(p `'LeftOuterJoin`` mb) -> do
+-- 'on' ('just' (p '^.' PersonId) '==.' mb '?.' BlogPostAuthorId)
+-- 'orderBy' ['asc' (p '^.' PersonName), 'asc' (mb '?.' BlogPostTitle)]
 -- return (p, mb)
 -- @
 --
@@ -310,10 +318,10 @@ import qualified Database.Persist
 -- In @esqueleto@, we may write the same query above as:
 --
 -- @
--- select $
--- from $ \\(p1 ``InnerJoin`` f ``InnerJoin`` p2) -> do
--- on (p2 ^. PersonId ==. f ^. FollowFollowed)
--- on (p1 ^. PersonId ==. f ^. FollowFollower)
+-- 'select' $
+-- 'from' $ \\(p1 `'InnerJoin`` f `'InnerJoin`` p2) -> do
+-- 'on' (p2 '^.' PersonId '==.' f '^.' FollowFollowed)
+-- 'on' (p1 '^.' PersonId '==.' f '^.' FollowFollower)
 -- return (p1, f, p2)
 -- @
 --
@@ -327,11 +335,11 @@ import qualified Database.Persist
 --
 -- @
 -- do 'update' $ \\p -> do
---      'set' p [ PersonName '=.' val \"João\" ]
---      where_ (p ^. PersonName ==. val \"Joao\")
+--      'set' p [ PersonName '=.' 'val' \"João\" ]
+--      'where_' (p '^.' PersonName '==.' 'val' \"Joao\")
 --    'delete' $
---      from $ \\p -> do
---      where_ (p ^. PersonAge <. just (val 14))
+--      'from' $ \\p -> do
+--      'where_' (p '^.' PersonAge '<.' 'just' ('val' 14))
 -- @
 --
 -- The results of queries can also be used for insertions.
@@ -347,8 +355,8 @@ import qualified Database.Persist
 -- In @esqueleto@, we may write the same query above as:
 --
 -- @
--- insertSelect $ from $ \\p->
---  return $ BlogPost \<# \"Group Blog Post\" \<&\> (p ^. PersonId)
+--  'insertSelect' $ 'from' $ \\p->
+--  return $ BlogPost '<#' \"Group Blog Post\" '<&>' (p '^.' PersonId)
 -- @
 --
 -- Individual insertions can be performed through Persistent's
@@ -377,8 +385,24 @@ import qualified Database.Persist
 ----------------------------------------------------------------------
 
 
--- | @valkey i = val (Key (PersistInt64 i))@
--- (<https://github.com/meteficha/esqueleto/issues/9>).
+-- $rdbmsSpecificModules
+--
+-- There are many differences between SQL syntax and functions
+-- supported by different RDBMSs.  Since version 2.2.8,
+-- @esqueleto@ includes modules containing functions that are
+-- specific to a given RDBMS.
+--
+--  * PostgreSQL: "Database.Esqueleto.PostgreSQL".
+--
+-- In order to use these functions, you need to explicitly import
+-- their corresponding modules, they're not re-exported here.
+
+
+----------------------------------------------------------------------
+
+
+-- | @valkey i = 'val' . 'toSqlKey'@
+-- (<https://github.com/prowdsponsor/esqueleto/issues/9>).
 valkey :: (Esqueleto query expr backend, ToBackendKey SqlBackend entity, PersistField (Key entity)) =>
           Int64 -> expr (Value (Key entity))
 valkey = val . toSqlKey
