@@ -27,8 +27,11 @@ module Database.Esqueleto.Internal.Language
   , FullOuterJoin(..)
   , OnClauseWithoutMatchingJoinException(..)
   , OrderBy
+  , DistinctOn
   , Update
   , Insertion
+  , LockingKind(..)
+  , SqlString
     -- * The guts
   , JoinKind(..)
   , IsJoinKind(..)
@@ -43,9 +46,14 @@ module Database.Esqueleto.Internal.Language
 import Control.Applicative (Applicative(..), (<$>))
 import Control.Exception (Exception)
 import Data.Int (Int64)
-import Data.String (IsString)
 import Data.Typeable (Typeable)
 import Database.Esqueleto.Internal.PersistentImport
+import Text.Blaze.Html (Html)
+
+import qualified Data.ByteString as B
+import qualified Data.ByteString.Lazy as BL
+import qualified Data.Text as T
+import qualified Data.Text.Lazy as TL
 
 
 -- | Finally tagless representation of @esqueleto@'s EDSL.
@@ -99,8 +107,8 @@ class (Functor query, Applicative query, Monad query) =>
   --
   -- @
   -- select $
-  -- from $ \\(foo ``InnerJoin`` bar) -> do
-  --   on (foo ^. FooId ==. bar ^. BarFooId)
+  -- 'from' $ \\(foo `'InnerJoin`` bar) -> do
+  --   'on' (foo '^.' FooId '==.' bar '^.' BarFooId)
   --   ...
   -- @
   --
@@ -111,9 +119,9 @@ class (Functor query, Applicative query, Monad query) =>
   --
   -- @
   -- select $
-  -- from $ \\(foo ``InnerJoin`` bar ``InnerJoin`` baz) -> do
-  --   on (baz ^. BazId ==. bar ^. BarBazId)
-  --   on (foo ^. FooId ==. bar ^. BarFooId)
+  -- 'from' $ \\(foo `'InnerJoin`` bar `'InnerJoin`` baz) -> do
+  --   'on' (baz '^.' BazId '==.' bar '^.' BarBazId)
+  --   'on' (foo '^.' FooId '==.' bar '^.' BarFooId)
   --   ...
   -- @
   --
@@ -122,11 +130,11 @@ class (Functor query, Applicative query, Monad query) =>
   --
   -- @
   -- let query1 =
-  --       from $ \\(foo ``InnerJoin`` bar) -> do
-  --         on (foo ^. FooId ==. bar ^. BarFooId)
+  --       'from' $ \\(foo `'InnerJoin`` bar) -> do
+  --         'on' (foo '^.' FooId '==.' bar '^.' BarFooId)
   --     query2 =
-  --       from $ \\(mbaz ``LeftOuterJoin`` quux) -> do
-  --         return (mbaz ?. BazName, quux)
+  --       'from' $ \\(mbaz `'LeftOuterJoin`` quux) -> do
+  --         return (mbaz '?.' BazName, quux)
   --     test1 =      (,) \<$\> query1 \<*\> query2
   --     test2 = flip (,) \<$\> query2 \<*\> query1
   -- @
@@ -140,30 +148,34 @@ class (Functor query, Applicative query, Monad query) =>
   -- in a tuple.
   --
   -- @
-  -- select $ from \\(foo ``InnerJoin`` bar) -> do
-  --   on (foo ^. FooBarId ==. bar ^. BarId)
-  --   groupBy (bar ^. BarId, bar ^. BarName)
-  --   return (bar ^. BarId, bar ^. BarName, countRows)
+  -- select $ 'from' \\(foo `'InnerJoin`` bar) -> do
+  --   'on' (foo '^.' FooBarId '==.' bar '^.' BarId)
+  --   'groupBy' (bar '^.' BarId, bar '^.' BarName)
+  --   return (bar '^.' BarId, bar '^.' BarName, countRows)
   -- @
   --
-  -- With groupBy you can sort by aggregate functions, like so (we
-  -- used @let@ to restrict the more general `countRows` to
-  -- @SqlExpr (Value Int)@ to avoid ambiguity):
+  -- With groupBy you can sort by aggregate functions, like so
+  -- (we used @let@ to restrict the more general 'countRows' to
+  -- @SqlExpr (Value Int)@ to avoid ambiguity---the second use of
+  -- 'countRows' has its type restricted by the @:: Int@ below):
   --
   -- @
-  -- r \<- select $ from \\(foo ``InnerJoin`` bar) -> do
-  --   on (foo ^. FooBarId ==. bar ^. BarId)
-  --   groupBy $ bar ^. BarName
-  --   let countRows' = countRows
-  --   orderBy [asc countRows']
-  --   return (bar ^. BarName, countRows')
-  -- forM_ r $ \\((Value name), (Value count)) -> do
+  -- r \<- select $ 'from' \\(foo `'InnerJoin`` bar) -> do
+  --   'on' (foo '^.' FooBarId '==.' bar '^.' BarId)
+  --   'groupBy' $ bar '^.' BarName
+  --   let countRows' = 'countRows'
+  --   'orderBy' ['asc' countRows']
+  --   return (bar '^.' BarName, countRows')
+  -- forM_ r $ \\('Value' name, 'Value' count) -> do
   --   print name
   --   print (count :: Int)
   -- @
   groupBy :: (ToSomeValues expr a) => a -> query ()
 
   -- | @ORDER BY@ clause. See also 'asc' and 'desc'.
+  --
+  -- Multiple calls to 'orderBy' get concatenated on the final
+  -- query, including 'distinctOnOrderBy'.
   orderBy :: [expr OrderBy] -> query ()
 
   -- | Ascending order of this field or expression.
@@ -178,6 +190,86 @@ class (Functor query, Applicative query, Monad query) =>
   -- | @OFFSET@.  Usually used with 'limit'.
   offset :: Int64 -> query ()
 
+  -- | @DISTINCT@.  Change the current @SELECT@ into @SELECT
+  -- DISTINCT@.  For example:
+  --
+  -- @
+  -- select $ distinct $
+  --   'from' \\foo -> do
+  --   ...
+  -- @
+  --
+  -- Note that this also has the same effect:
+  --
+  -- @
+  -- select $
+  --   'from' \\foo -> do
+  --   distinct (return ())
+  --   ...
+  -- @
+  --
+  -- /Since: 2.2.4/
+  distinct :: query a -> query a
+
+  -- | @DISTINCT ON@.  Change the current @SELECT@ into
+  -- @SELECT DISTINCT ON (expressions)@.  For example:
+  --
+  -- @
+  -- select $
+  --   'from' \\foo ->
+  --   'distinctOn' ['don' (foo ^. FooName), 'don' (foo ^. FooState)] $ do
+  --   ...
+  -- @
+  --
+  -- You can also chain different calls to 'distinctOn'.  The
+  -- above is equivalent to:
+  --
+  -- @
+  -- select $
+  --   'from' \\foo ->
+  --   'distinctOn' ['don' (foo ^. FooName)] $
+  --   'distinctOn' ['don' (foo ^. FooState)] $ do
+  --   ...
+  -- @
+  --
+  -- Each call to 'distinctOn' adds more expressions.  Calls to
+  -- 'distinctOn' override any calls to 'distinct'.
+  --
+  -- Note that PostgreSQL requires the expressions on @DISTINCT
+  -- ON@ to be the first ones to appear on a @ORDER BY@.  This is
+  -- not managed automatically by esqueleto, keeping its spirit
+  -- of trying to be close to raw SQL.
+  --
+  -- Supported by PostgreSQL only.
+  --
+  -- /Since: 2.2.4/
+  distinctOn :: [expr DistinctOn] -> query a -> query a
+
+  -- | Erase an expression's type so that it's suitable to
+  -- be used by 'distinctOn'.
+  --
+  -- /Since: 2.2.4/
+  don :: expr (Value a) -> expr DistinctOn
+
+  -- | A convenience function that calls both 'distinctOn' and
+  -- 'orderBy'.  In other words,
+  --
+  -- @
+  -- 'distinctOnOrderBy' [asc foo, desc bar, desc quux] $ do
+  --   ...
+  -- @
+  --
+  -- is the same as:
+  --
+  -- @
+  -- 'distinctOn' [don foo, don  bar, don  quux] $ do
+  --   'orderBy'  [asc foo, desc bar, desc quux]
+  --   ...
+  -- @
+  --
+  -- /Since: 2.2.4/
+  distinctOnOrderBy :: [expr OrderBy] -> query a -> query a
+
   -- | @ORDER BY random()@ clause.
   --
   -- /Since: 1.3.10/
@@ -188,6 +280,15 @@ class (Functor query, Applicative query, Monad query) =>
   -- /Since: 1.2.2/
   having :: expr (Value Bool) -> query ()
 
+  -- | Add a locking clause to the query.  Please read
+  -- 'LockingKind' documentation and your RDBMS manual.
+  --
+  -- If multiple calls to 'locking' are made on the same query,
+  -- the last one is used.
+  --
+  -- /Since: 2.2.7/
+  locking :: LockingKind -> query ()
+
   -- | Execute a subquery @SELECT@ in an expression.  Returns a
   -- simple value so should be used only when the @SELECT@ query
   -- is guaranteed to return just one row.
@@ -211,8 +312,8 @@ class (Functor query, Applicative query, Monad query) =>
   isNothing :: PersistField typ => expr (Value (Maybe typ)) -> expr (Value Bool)
 
   -- | Analogous to 'Just', promotes a value of type @typ@ into
-  -- one of type @Maybe typ@.  It should hold that @val . Just
-  -- === just . val@.
+  -- one of type @Maybe typ@.  It should hold that @'val' . Just
+  -- === just . 'val'@.
   just :: expr (Value typ) -> expr (Value (Maybe typ))
 
   -- | @NULL@ value.
@@ -226,7 +327,12 @@ class (Functor query, Applicative query, Monad query) =>
   countRows :: Num a => expr (Value a)
 
   -- | @COUNT@.
-  count :: (Num a) => expr (Value typ) -> expr (Value a)
+  count :: Num a => expr (Value typ) -> expr (Value a)
+
+  -- | @COUNT(DISTINCT x)@.
+  --
+  -- /Since: 2.4.1/
+  countDistinct :: Num a => expr (Value typ) -> expr (Value a)
 
   not_ :: expr (Value Bool) -> expr (Value Bool)
 
@@ -255,7 +361,29 @@ class (Functor query, Applicative query, Monad query) =>
   min_     :: (PersistField a) => expr (Value a) -> expr (Value (Maybe a))
   max_     :: (PersistField a) => expr (Value a) -> expr (Value (Maybe a))
   avg_     :: (PersistField a, PersistField b) => expr (Value a) -> expr (Value (Maybe b))
-  lower_   :: (PersistField a, IsString a) => expr (Value a) -> expr (Value a)
+
+  -- | Allow a number of one type to be used as one of another
+  -- type via an implicit cast.  An explicit cast is not made,
+  -- this function changes only the types on the Haskell side.
+  --
+  -- /Caveat/: Trying to use @castNum@ from @Double@ to @Int@
+  -- will not result in an integer, the original fractional
+  -- number will still be used!  Use 'round_', 'ceiling_' or
+  -- 'floor_' instead.
+  --
+  -- /Safety/: This operation is mostly safe due to the 'Num'
+  -- constraint between the types and the fact that RDBMSs
+  -- usually allow numbers of different types to be used
+  -- interchangeably.  However, there may still be issues with
+  -- the query not being accepted by the RDBMS or @persistent@
+  -- not being able to parse it.
+  --
+  -- /Since: 2.2.9/
+  castNum :: (Num a, Num b) => expr (Value a) -> expr (Value b)
+  -- | Same as 'castNum', but for nullable values.
+  --
+  -- /Since: 2.2.9/
+  castNumM :: (Num a, Num b) => expr (Value (Maybe a)) -> expr (Value (Maybe b))
 
   -- | @COALESCE@ function. Evaluates the arguments in order and
   -- returns the value of the first non-NULL expression, or NULL
@@ -273,24 +401,43 @@ class (Functor query, Applicative query, Monad query) =>
   -- /Since: 1.4.3/
   coalesceDefault :: PersistField a => [expr (Value (Maybe a))] -> expr (Value a) -> expr (Value a)
 
+  -- | @LOWER@ function.
+  lower_ :: SqlString s => expr (Value s) -> expr (Value s)
   -- | @LIKE@ operator.
-  like :: (PersistField s, IsString s) => expr (Value s) -> expr (Value s) -> expr (Value Bool)
+  like :: SqlString s => expr (Value s) -> expr (Value s) -> expr (Value Bool)
+  -- | @ILIKE@ operator (case-insensitive @LIKE@).
+  --
+  -- Supported by PostgreSQL only.
+  --
+  -- /Since: 2.2.3/
+  ilike :: SqlString s => expr (Value s) -> expr (Value s) -> expr (Value Bool)
   -- | The string @'%'@.  May be useful while using 'like' and
   -- concatenation ('concat_' or '++.', depending on your
   -- database).  Note that you always to type the parenthesis,
   -- for example:
   --
   -- @
-  -- name ``'like'`` (%) ++. val "John" ++. (%)
+  -- name `'like`` (%) ++. 'val' \"John\" ++. (%)
   -- @
-  (%) :: (PersistField s, IsString s) => expr (Value s)
+  (%) :: SqlString s => expr (Value s)
   -- | The @CONCAT@ function with a variable number of
   -- parameters.  Supported by MySQL and PostgreSQL.
-  concat_ :: (PersistField s, IsString s) => [expr (Value s)] -> expr (Value s)
+  concat_ :: SqlString s => [expr (Value s)] -> expr (Value s)
   -- | The @||@ string concatenation operator (named after
   -- Haskell's '++' in order to avoid naming clash with '||.').
   -- Supported by SQLite and PostgreSQL.
-  (++.) :: (PersistField s, IsString s) => expr (Value s) -> expr (Value s) -> expr (Value s)
+  (++.) :: SqlString s => expr (Value s) -> expr (Value s) -> expr (Value s)
+  -- | Cast a string type into 'Text'.  This function
+  -- is very useful if you want to use @newtype@s, or if you want
+  -- to apply functions such as 'like' to strings of different
+  -- types.
+  --
+  -- /Safety:/ This is a slightly unsafe function, especially if
+  -- you have defined your own instances of 'SqlString'.  Also,
+  -- since 'Maybe' is an instance of 'SqlString', it's possible
+  -- to turn a nullable value into a non-nullable one.  Avoid
+  -- using this function if possible.
+  castString :: (SqlString s, SqlString r) => expr (Value s) -> expr (Value r)
 
   -- | Execute a subquery @SELECT@ in an expression.  Returns a
   -- list of values.
@@ -302,6 +449,13 @@ class (Functor query, Applicative query, Monad query) =>
   -- | Lift a list of constant value from Haskell-land to the query.
   valList :: PersistField typ => [typ] -> expr (ValueList typ)
 
+  -- | Same as 'just' but for 'ValueList'.  Most of the time you
+  -- won't need it, though, because you can use 'just' from
+  -- inside 'subList_select' or 'Just' from inside 'valList'.
+  --
+  -- /Since: 2.2.12/
+  justList :: expr (ValueList typ) -> expr (ValueList (Maybe typ))
+
   -- | @IN@ operator.
   in_ :: PersistField typ => expr (Value typ) -> expr (ValueList typ) -> expr (Value Bool)
 
@@ -312,10 +466,10 @@ class (Functor query, Applicative query, Monad query) =>
   --
   -- @
   -- select $
-  -- from $ \\person -> do
-  -- where_ $ exists $
-  --          from $ \\post -> do
-  --          where_ (post ^. BlogPostAuthorId ==. person ^. PersonId)
+  -- 'from' $ \\person -> do
+  -- 'where_' $ 'exists' $
+  --          'from' $ \\post -> do
+  --          'where_' (post '^.' BlogPostAuthorId '==.' person '^.' PersonId)
   -- return person
   -- @
   exists :: query () -> expr (Value Bool)
@@ -345,31 +499,31 @@ class (Functor query, Applicative query, Monad query) =>
   -- @
   -- select $
   -- return $
-  -- case_
-  --    [ when_
-  --        (exists $
-  --        from $ \\p -> do
-  --        where_ (p ^. PersonName ==. val "Mike"))
-  --      then_
-  --        (sub_select $
-  --        from $ \\v -> do
+  -- 'case_'
+  --    [ 'when_'
+  --        ('exists' $
+  --        'from' $ \\p -> do
+  --        'where_' (p '^.' PersonName '==.' 'val' \"Mike\"))
+  --      'then_'
+  --        ('sub_select' $
+  --        'from' $ \\v -> do
   --        let sub =
-  --                from $ \\c -> do
-  --                where_ (c ^. PersonName ==. val "Mike")
-  --                return (c ^. PersonFavNum)
-  --        where_ (v ^. PersonFavNum >. sub_select sub)
-  --        return $ count (v ^. PersonName) +. val (1 :: Int)) ]
-  --    (else_ $ val (-1))
+  --                'from' $ \\c -> do
+  --                'where_' (c '^.' PersonName '==.' 'val' \"Mike\")
+  --                return (c '^.' PersonFavNum)
+  --        'where_' (v '^.' PersonFavNum >. 'sub_select' sub)
+  --        return $ 'count' (v '^.' PersonName) +. 'val' (1 :: Int)) ]
+  --    ('else_' $ 'val' (-1))
   -- @
   --
   -- This query is a bit complicated, but basically it checks if a person
-  -- named "Mike" exists, and if that person does, run the subquery to find
-  -- out how many people have a ranking (by Fav Num) higher than "Mike".
+  -- named @\"Mike\"@ exists, and if that person does, run the subquery to find
+  -- out how many people have a ranking (by Fav Num) higher than @\"Mike\"@.
   --
   -- __NOTE:__ There are a few things to be aware about this statement.
   --
   --    * This only implements the full CASE statement, it does not
-  --      implement the "simple" CASE statement.
+  --      implement the \"simple\" CASE statement.
   --
   --
   --    * At least one 'when_' and 'then_' is mandatory otherwise it will
@@ -383,6 +537,10 @@ class (Functor query, Applicative query, Monad query) =>
   -- /Since: 2.1.2/
   case_ :: PersistField a => [(expr (Value Bool), expr (Value a))] -> expr (Value a) -> expr (Value a)
 
+{-# DEPRECATED sub_selectDistinct "Since 2.2.4: use 'sub_select' and 'distinct'." #-}
+{-# DEPRECATED subList_selectDistinct "Since 2.2.4: use 'subList_select' and 'distinct'." #-}
+
+
 -- Fixity declarations
 infixl 9 ^.
 infixl 7 *., /.
@@ -390,7 +548,8 @@ infixl 6 +., -.
 infixr 5 ++.
 infix  4 ==., >=., >., <=., <., !=.
 infixr 3 &&., =., +=., -=., *=., /=.
-infixr 2 ||., `InnerJoin`, `CrossJoin`, `LeftOuterJoin`, `RightOuterJoin`, `FullOuterJoin`, `like`
+infixr 2 ||., `like`, `ilike`
+infixl 2 `InnerJoin`, `CrossJoin`, `LeftOuterJoin`, `RightOuterJoin`, `FullOuterJoin`
 
 -- | Syntax sugar for 'case_'.
 --
@@ -442,8 +601,8 @@ data SomeValue expr where
   SomeValue :: Esqueleto query expr backend => expr (Value a) -> SomeValue expr
 
 -- | A class of things that can be converted into a list of SomeValue. It has
--- instances for tuples and is the reason why groupBy can take tuples, like
--- @groupBy (foo ^. FooId, foo ^. FooName, foo ^. FooType)@.
+-- instances for tuples and is the reason why 'groupBy' can take tuples, like
+-- @'groupBy' (foo '^.' FooId, foo '^.' FooName, foo '^.' FooType)@.
 class ToSomeValues expr a where
   toSomeValues :: a -> [SomeValue expr]
 
@@ -521,7 +680,7 @@ data CrossJoin a b = a `CrossJoin` b
 --
 -- @
 -- select $
--- from $ \\(person ``LeftOuterJoin`` pet) ->
+-- 'from' $ \\(person `'LeftOuterJoin`` pet) ->
 --   ...
 -- @
 --
@@ -550,6 +709,7 @@ data JoinKind =
   | LeftOuterJoinKind  -- ^ @LEFT OUTER JOIN@
   | RightOuterJoinKind -- ^ @RIGHT OUTER JOIN@
   | FullOuterJoinKind  -- ^ @FULL OUTER JOIN@
+    deriving Eq
 
 
 -- | (Internal) Functions that operate on types (that should be)
@@ -593,6 +753,10 @@ data PreprocessedFrom a
 data OrderBy
 
 
+-- | Phantom type used by 'distinctOn' and 'don'.
+data DistinctOn
+
+
 -- | Phantom type for a @SET@ operation on an entity of the given
 -- type (see 'set' and '(=.)').
 data Update typ
@@ -602,6 +766,60 @@ data Update typ
 data Insertion a
 
 
+-- | Different kinds of locking clauses supported by 'locking'.
+--
+-- Note that each RDBMS has different locking support.  The
+-- constructors of this datatype specify only the /syntax/ of the
+-- locking mechanism, not its /semantics/.  For example, even
+-- though both MySQL and PostgreSQL support 'ForUpdate', there
+-- are no guarantees that they will behave the same.
+--
+-- /Since: 2.2.7/
+data LockingKind =
+    ForUpdate
+    -- ^ @FOR UPDATE@ syntax.  Supported by MySQL, Oracle and
+    -- PostgreSQL.
+    --
+    -- /Since: 2.2.7/
+  | ForShare
+    -- ^ @FOR SHARE@ syntax.  Supported by PostgreSQL.
+    --
+    -- /Since: 2.2.7/
+  | LockInShareMode
+    -- ^ @LOCK IN SHARE MODE@ syntax.  Supported by MySQL.
+    --
+    -- /Since: 2.2.7/
+
+
+-- | Phantom class of data types that are treated as strings by the
+-- RDBMS.  It has no methods because it's only used to avoid type
+-- errors such as trying to concatenate integers.
+--
+-- If you have a custom data type or @newtype@, feel free to make
+-- it an instance of this class.
+--
+-- /Since: 2.4.0/
+class PersistField a => SqlString a where
+
+-- | /Since: 2.3.0/
+instance (a ~ Char) => SqlString [a] where
+
+-- | /Since: 2.3.0/
+instance SqlString T.Text where
+
+-- | /Since: 2.3.0/
+instance SqlString TL.Text where
+
+-- | /Since: 2.3.0/
+instance SqlString B.ByteString where
+
+-- | /Since: 2.3.0/
+instance SqlString Html where
+
+-- | /Since: 2.4.0/
+instance SqlString a => SqlString (Maybe a) where
+
+
 -- | @FROM@ clause: bring entities into scope.
 --
 -- This function internally uses two type classes in order to
@@ -620,7 +838,7 @@ data Insertion a
 --  * A @JOIN@ of any other two types allowed by the innermost
 --  magic, where a @JOIN@ may be an 'InnerJoin', a 'CrossJoin', a
 --  'LeftOuterJoin', a 'RightOuterJoin', or a 'FullOuterJoin'.
---  The @JOINs@ have right fixity, the same as in SQL.
+--  The @JOINs@ have left fixity.
 --
 -- The outermost magic allows you to use @from@ on any tuples of
 -- types supported by innermost magic (and also tuples of tuples,
@@ -634,11 +852,11 @@ data Insertion a
 -- of the arguments of the lambda are inside square brackets):
 --
 -- @
--- from $ \\person -> ...
--- from $ \\(person, blogPost) -> ...
--- from $ \\(p ``LeftOuterJoin`` mb) -> ...
--- from $ \\(p1 ``InnerJoin`` f ``InnerJoin`` p2) -> ...
--- from $ \\((p1 ``InnerJoin`` f) ``InnerJoin`` p2) -> ...
+-- 'from' $ \\person -> ...
+-- 'from' $ \\(person, blogPost) -> ...
+-- 'from' $ \\(p `'LeftOuterJoin`` mb) -> ...
+-- 'from' $ \\(p1 `'InnerJoin`` f `'InnerJoin`` p2) -> ...
+-- 'from' $ \\((p1 `'InnerJoin`` f) `'InnerJoin`` p2) -> ...
 -- @
 --
 -- The types of the arguments to the lambdas above are,
@@ -652,25 +870,21 @@ data Insertion a
 --      ) => expr (Entity Person)
 -- (person, blogPost)
 --   :: (...) => (expr (Entity Person), expr (Entity BlogPost))
--- (p ``LeftOuterJoin`` mb)
+-- (p `'LeftOuterJoin`` mb)
 --   :: (...) => InnerJoin (expr (Entity Person)) (expr (Maybe (Entity BlogPost)))
--- (p1 ``InnerJoin`` f ``InnerJoin`` p2)
---   :: (...) => InnerJoin
---                 (expr (Entity Person))
---                 (InnerJoin (expr (Entity Follow))
---                            (expr (Entity Person)))
--- ((p1 ``InnerJoin`` f) ``InnerJoin`` p2) ::
+-- (p1 `'InnerJoin`` f `'InnerJoin`` p2)
 --   :: (...) => InnerJoin
 --                 (InnerJoin (expr (Entity Person))
 --                            (expr (Entity Follow)))
 --                 (expr (Entity Person))
+-- (p1 `'InnerJoin`` (f `'InnerJoin`` p2)) ::
+--   :: (...) => InnerJoin
+--                 (expr (Entity Person))
+--                 (InnerJoin (expr (Entity Follow))
+--                            (expr (Entity Person)))
 -- @
 --
 -- Note that some backends may not support all kinds of @JOIN@s.
--- For example, when using the SQL backend with SQLite, it will
--- not accept the last example above (which is associated to the
--- left, instead of being to the right) and will not accept
--- 'RightOuterJoin's or 'FullOuterJoin's.
 from :: From query expr backend a => (a -> query b) -> query b
 from = (from_ >>=)
 
