@@ -31,7 +31,7 @@ module Database.Esqueleto.Internal.Sql
     -- * The guts
   , unsafeSqlCase
   , unsafeSqlBinOp
-  , unsafeSqlBinOpList
+  , unsafeSqlBinOpComposite
   , unsafeSqlValue
   , unsafeSqlFunction
   , unsafeSqlExtractSubField
@@ -55,15 +55,15 @@ import Control.Exception (throw, throwIO)
 import Control.Monad (ap, MonadPlus(..), liftM)
 import Control.Monad.IO.Class (MonadIO(..))
 import Control.Monad.Trans.Class (lift)
-import qualified Control.Monad.Trans.Reader as R
+import Control.Monad.Trans.Resource (MonadResource)
+import Data.Acquire (with, allocateAcquire, Acquire)
 import Data.Int (Int64)
 import Data.List (intersperse)
-import Data.Monoid (Monoid(..), (<>))
+import Data.Monoid (Last(..), Monoid(..), (<>))
 import Data.Proxy (Proxy(..))
 import Database.Esqueleto.Internal.PersistentImport
-import Database.Persist.Sql.Util (
-    entityColumnNames, entityColumnCount, parseEntityValues, isIdField
-  , hasCompositeKey)
+import Database.Persist.Sql.Util (entityColumnNames, entityColumnCount, parseEntityValues, isIdField, hasCompositeKey)
+import qualified Control.Monad.Trans.Reader as R
 import qualified Control.Monad.Trans.State as S
 import qualified Control.Monad.Trans.Writer as W
 import qualified Data.Conduit as C
@@ -72,8 +72,6 @@ import qualified Data.HashSet as HS
 import qualified Data.Text as T
 import qualified Data.Text.Lazy as TL
 import qualified Data.Text.Lazy.Builder as TLB
-import Data.Acquire (with, allocateAcquire, Acquire)
-import Control.Monad.Trans.Resource (MonadResource)
 
 import Database.Esqueleto.Internal.Language
 
@@ -103,19 +101,35 @@ type SqlEntity ent = (PersistEntity ent, PersistEntityBackend ent ~ SqlBackend)
 
 
 -- | Side data written by 'SqlQuery'.
-data SideData = SideData { sdFromClause    :: ![FromClause]
-                         , sdSetClause     :: ![SetClause]
-                         , sdWhereClause   :: !WhereClause
-                         , sdGroupByClause :: !GroupByClause
-                         , sdHavingClause  :: !HavingClause
-                         , sdOrderByClause :: ![OrderByClause]
-                         , sdLimitClause   :: !LimitClause
+data SideData = SideData { sdDistinctClause :: !DistinctClause
+                         , sdFromClause     :: ![FromClause]
+                         , sdSetClause      :: ![SetClause]
+                         , sdWhereClause    :: !WhereClause
+                         , sdGroupByClause  :: !GroupByClause
+                         , sdHavingClause   :: !HavingClause
+                         , sdOrderByClause  :: ![OrderByClause]
+                         , sdLimitClause    :: !LimitClause
+                         , sdLockingClause  :: !LockingClause
                          }
 
 instance Monoid SideData where
-  mempty = SideData mempty mempty mempty mempty mempty mempty mempty
-  SideData f s w g h o l `mappend` SideData f' s' w' g' h' o' l' =
-    SideData (f <> f') (s <> s') (w <> w') (g <> g') (h <> h') (o <> o') (l <> l')
+  mempty = SideData mempty mempty mempty mempty mempty mempty mempty mempty mempty
+  SideData d f s w g h o l k `mappend` SideData d' f' s' w' g' h' o' l' k' =
+    SideData (d <> d') (f <> f') (s <> s') (w <> w') (g <> g') (h <> h') (o <> o') (l <> l') (k <> k')
+
+
+-- | The @DISTINCT@ "clause".
+data DistinctClause =
+    DistinctAll                     -- ^ The default, everything.
+  | DistinctStandard                -- ^ Only @DISTINCT@, SQL standard.
+  | DistinctOn [SqlExpr DistinctOn] -- ^ @DISTINCT ON@, PostgreSQL extension.
+
+instance Monoid DistinctClause where
+  mempty = DistinctAll
+  DistinctOn a     `mappend` DistinctOn b = DistinctOn (a <> b)
+  DistinctOn a     `mappend` _            = DistinctOn a
+  DistinctStandard `mappend` _            = DistinctStandard
+  DistinctAll      `mappend` b            = b
 
 
 -- | A part of a @FROM@ clause.
@@ -152,7 +166,9 @@ collectOnClauses = go []
           matchR = (\r' -> FromJoin l k r' onClause) <$> tryMatch expr r
           matchL = (\l' -> FromJoin l' k r onClause) <$> tryMatch expr l
           matchC = case onClause of
-                     Nothing -> return (FromJoin l k r (Just expr))
+                     Nothing | k /= CrossJoinKind
+                               -> return (FromJoin l k r (Just expr))
+                             | otherwise -> mzero
                      Just _  -> mzero
     tryMatch _ _ = mzero
 
@@ -194,6 +210,10 @@ instance Monoid LimitClause where
     -- "reversed" arguments.
 
 
+-- | A locking clause.
+type LockingClause = Last LockingKind
+
+
 ----------------------------------------------------------------------
 
 
@@ -244,6 +264,9 @@ useIdent info (I ident) = fromDBName info $ DBName ident
 
 
 -- | An expression on the SQL backend.
+--
+-- There are many comments describing the constructors of this
+-- data type.  However, Haddock doesn't like GADTs, so you'll have to read them by hitting \"Source\".
 data SqlExpr a where
   -- An entity, created by 'from' (cf. 'fromStart').
   EEntity  :: Ident -> SqlExpr (Entity val)
@@ -258,8 +281,47 @@ data SqlExpr a where
   -- interpolated by the SQL backend.
   ERaw     :: NeedParens -> (IdentInfo -> (TLB.Builder, [PersistValue])) -> SqlExpr (Value a)
 
-  -- Used to support non-id/composite primary keys
-  ERawList :: (IdentInfo -> ([TLB.Builder], [PersistValue])) -> SqlExpr (Value a)
+  -- A composite key.
+  --
+  -- Persistent uses the same 'PersistList' constructor for both
+  -- fields which are (homogeneous) lists of values and the
+  -- (probably heterogeneous) values of a composite primary key.
+  --
+  -- We need to treat composite keys as fields.  For example, we
+  -- have to support using ==., otherwise you wouldn't be able to
+  -- join.  OTOH, lists of values should be treated exactly the
+  -- same as any other scalar value.
+  --
+  -- In particular, this is valid for persistent via rawSql for
+  -- an F field that is a list:
+  --
+  --   A.F in ?    -- [PersistList [foo, bar]]
+  --
+  -- However, this is not for a composite key entity:
+  --
+  --   A.ID = ?    -- [PersistList [foo, bar]]
+  --
+  -- The ID field doesn't exist on the DB for a composite key
+  -- table, it exists only on the Haskell side.  Those variations
+  -- also don't work:
+  --
+  --   (A.KeyA, A.KeyB) = ?    -- [PersistList [foo, bar]]
+  --   [A.KeyA, A.KeyB] = ?    -- [PersistList [foo, bar]]
+  --
+  -- We have to generate:
+  --
+  --   A.KeyA = ? AND A.KeyB = ?      -- [foo, bar]
+  --
+  -- Note that the PersistList had to be deconstructed into its
+  -- components.
+  --
+  -- In order to disambiguate behaviors, this constructor is used
+  -- /only/ to represent a composite field access.  It does not
+  -- represent a 'PersistList', not even if the 'PersistList' is
+  -- used in the context of a composite key.  That's because it's
+  -- impossible, e.g., for 'val' to disambiguate between these
+  -- uses.
+  ECompositeKey :: (IdentInfo -> [TLB.Builder]) -> SqlExpr (Value a)
 
   -- 'EList' and 'EEmptyList' are used by list operators.
   EList      :: SqlExpr (Value a) -> SqlExpr (ValueList a)
@@ -269,6 +331,9 @@ data SqlExpr a where
   EOrderBy :: OrderByType -> SqlExpr (Value a) -> SqlExpr OrderBy
   EOrderRandom :: SqlExpr OrderBy
 
+  -- A 'SqlExpr' accepted only by 'distinctOn'.
+  EDistinctOn :: SqlExpr (Value a) -> SqlExpr DistinctOn
+
   -- A 'SqlExpr' accepted only by 'set'.
   ESet :: (SqlExpr (Entity val) -> SqlExpr (Value ())) -> SqlExpr (Update val)
 
@@ -330,6 +395,8 @@ instance Esqueleto SqlQuery SqlExpr SqlBackend where
 
   having expr = Q $ W.tell mempty { sdHavingClause = Where expr }
 
+  locking kind = Q $ W.tell mempty { sdLockingClause = Last (Just kind) }
+
   orderBy exprs = Q $ W.tell mempty { sdOrderByClause = exprs }
   asc  = EOrderBy ASC
   desc = EOrderBy DESC
@@ -339,49 +406,52 @@ instance Esqueleto SqlQuery SqlExpr SqlBackend where
   limit  n = Q $ W.tell mempty { sdLimitClause = Limit (Just n) Nothing  }
   offset n = Q $ W.tell mempty { sdLimitClause = Limit Nothing  (Just n) }
 
+  distinct         act = Q (W.tell mempty { sdDistinctClause = DistinctStandard }) >> act
+  distinctOn exprs act = Q (W.tell mempty { sdDistinctClause = DistinctOn exprs }) >> act
+  don = EDistinctOn
+  distinctOnOrderBy exprs act =
+    distinctOn (toDistinctOn <$> exprs) $ do
+      orderBy exprs
+      act
+    where
+      toDistinctOn :: SqlExpr OrderBy -> SqlExpr DistinctOn
+      toDistinctOn (EOrderBy _ f) = EDistinctOn f
+
   sub_select         = sub SELECT
-  sub_selectDistinct = sub SELECT_DISTINCT
+  sub_selectDistinct = sub_select . distinct
 
   (^.) :: forall val typ. (PersistEntity val, PersistField typ)
        => SqlExpr (Entity val) -> EntityField val typ -> SqlExpr (Value typ)
   EEntity ident ^. field
-    | isComposite = ERawList   $ \info -> (dot info <$> compositeFields pdef,  [])
-    | otherwise   = ERaw Never $ \info -> (dot info  $  persistFieldDef field, [])
+    | isComposite = ECompositeKey $ \info ->  dot info <$> compositeFields pdef
+    | otherwise   = ERaw Never    $ \info -> (dot info  $  persistFieldDef field, [])
     where
       isComposite = isIdField field && hasCompositeKey ed
       dot info x  = useIdent info ident <> "." <> fromDBName info (fieldDB x)
       ed          = entityDef $ getEntityVal (Proxy :: Proxy (SqlExpr (Entity val)))
       Just pdef   = entityPrimary ed
 
-  EMaybe r ?. field = maybelize (r ^. field)
-    where
-      maybelize :: SqlExpr (Value a) -> SqlExpr (Value (Maybe a))
-      maybelize (ERaw p f)   = ERaw p f
-      maybelize (ERawList f) = ERawList f
+  EMaybe r ?. field = just (r ^. field)
 
-  val v =
-    case toPersistValue v of
-      PersistList vs -> ERawList   $ const (replicate (length vs) "?", vs)
-      pv             -> ERaw Never $ const ("?", [pv])
+  val v = ERaw Never $ const ("?", [toPersistValue v])
 
-  isNothing (ERaw p f)   = ERaw Parens $ first ((<> " IS NULL") . parensM p) . f
-  isNothing (ERawList f) = ERaw Parens $ first (intersperseB " AND " . map (<> " IS NULL")) . f
-  just (ERaw p f)   = ERaw p f
-  just (ERawList f) = ERawList f
+  isNothing (ERaw p f)        = ERaw Parens $ first ((<> " IS NULL") . parensM p) . f
+  isNothing (ECompositeKey f) = ERaw Parens $ flip (,) [] . (intersperseB " AND " . map (<> " IS NULL")) . f
+  just (ERaw p f)        = ERaw p f
+  just (ECompositeKey f) = ECompositeKey f
   nothing = unsafeSqlValue "NULL"
-  joinV (ERaw p f)   = ERaw p f
-  joinV (ERawList f) = ERawList f
-  countRows = unsafeSqlValue "COUNT(*)"
-  count (ERaw _ f) = ERaw Never $ \info -> let (b, vals) = f info
-                                           in ("COUNT" <> parens b, vals)
-  count (ERawList _) = unsafeSqlValue "COUNT(*)" -- Assumes no NULLs on a PK
+  joinV (ERaw p f)        = ERaw p f
+  joinV (ECompositeKey f) = ECompositeKey f
+  countRows     = unsafeSqlValue "COUNT(*)"
+  count         = countHelper ""           ""
+  countDistinct = countHelper "(DISTINCT " ")"
 
   not_ (ERaw p f) = ERaw Never $ \info -> let (b, vals) = f info
                                           in ("NOT " <> parensM p b, vals)
-  not_ (ERawList _) = unexpectedCompositeKeyError
+  not_ (ECompositeKey _) = unexpectedCompositeKeyError "not_"
 
-  (==.) = unsafeSqlBinOpList " = " " AND "
-  (!=.) = unsafeSqlBinOpList " != " " OR "
+  (==.) = unsafeSqlBinOpComposite " = " " AND "
+  (!=.) = unsafeSqlBinOpComposite " != " " OR "
   (>=.) = unsafeSqlBinOp " >= "
   (>.)  = unsafeSqlBinOp " > "
   (<=.) = unsafeSqlBinOp " <= "
@@ -402,23 +472,31 @@ instance Esqueleto SqlQuery SqlExpr SqlBackend where
   min_     = unsafeSqlFunction "MIN"
   max_     = unsafeSqlFunction "MAX"
   avg_     = unsafeSqlFunction "AVG"
-  lower_   = unsafeSqlFunction "LOWER"
+
+  castNum  = veryUnsafeCoerceSqlExprValue
+  castNumM = veryUnsafeCoerceSqlExprValue
 
   coalesce              = unsafeSqlFunctionParens "COALESCE"
   coalesceDefault exprs = unsafeSqlFunctionParens "COALESCE" . (exprs ++) . return . just
 
+  lower_  = unsafeSqlFunction "LOWER"
   like    = unsafeSqlBinOp    " LIKE "
+  ilike   = unsafeSqlBinOp    " ILIKE "
   (%)     = unsafeSqlValue    "'%'"
   concat_ = unsafeSqlFunction "CONCAT"
   (++.)   = unsafeSqlBinOp    " || "
+  castString = veryUnsafeCoerceSqlExprValue
 
   subList_select         = EList . sub_select
-  subList_selectDistinct = EList . sub_selectDistinct
+  subList_selectDistinct = subList_select . distinct
 
   valList []   = EEmptyList
   valList vals = EList $ ERaw Parens $ const ( uncommas ("?" <$ vals)
                                              , map toPersistValue vals )
 
+  justList EEmptyList = EEmptyList
+  justList (EList v)  = EList (just v)
+
   v `in_`   e = ifNotEmptyList e False $ unsafeSqlBinOp     " IN " v (veryUnsafeCoerceSqlExprValueList e)
   v `notIn` e = ifNotEmptyList e True  $ unsafeSqlBinOp " NOT IN " v (veryUnsafeCoerceSqlExprValueList e)
 
@@ -435,14 +513,14 @@ instance Esqueleto SqlQuery SqlExpr SqlBackend where
   field *=. expr = setAux field (\ent -> ent ^. field *. expr)
   field /=. expr = setAux field (\ent -> ent ^. field /. expr)
 
-  (<#) _ (ERaw _ f)   = EInsert Proxy f
-  (<#) _ (ERawList _) = unexpectedCompositeKeyError
+  (<#) _ (ERaw _ f)        = EInsert Proxy f
+  (<#) _ (ECompositeKey _) = unexpectedCompositeKeyError "(<#)"
 
   (EInsert _ f) <&> (ERaw _ g) = EInsert Proxy $ \x ->
     let (fb, fv) = f x
         (gb, gv) = g x
     in (fb <> ", " <> gb, fv ++ gv)
-  (EInsert _ _) <&> (ERawList _) = unexpectedCompositeKeyError
+  (EInsert _ _) <&> (ECompositeKey _) = unexpectedCompositeKeyError "(<&>)"
 
   case_ = unsafeSqlCase
 
@@ -477,6 +555,10 @@ ifNotEmptyList :: SqlExpr (ValueList a) -> Bool -> SqlExpr (Value Bool) -> SqlEx
 ifNotEmptyList EEmptyList b _ = val b
 ifNotEmptyList (EList _)  _ x = x
 
+countHelper :: Num a => TLB.Builder -> TLB.Builder -> SqlExpr (Value typ) -> SqlExpr (Value a)
+countHelper open close (ERaw _ f) = ERaw Never $ first (\b -> "COUNT" <> open <> parens b <> close) . f
+countHelper _ _ (ECompositeKey _) = countRows -- Assumes no NULLs on a PK
+
 
 ----------------------------------------------------------------------
 
@@ -502,8 +584,8 @@ unsafeSqlCase when (ERaw p1 f1) = ERaw Never buildCase
         let (b1, vals1) = f1' info
             (b2, vals2) = f2 info
         in ( b0 <> " WHEN " <> parensM p1' b1 <> " THEN " <> parensM p2 b2, vals0 <> vals1 <> vals2 )
-    foldHelp _ _ _ = unexpectedCompositeKeyError
-unsafeSqlCase _ (ERawList _) = unexpectedCompositeKeyError
+    foldHelp _ _ _ = unexpectedCompositeKeyError "unsafeSqlCase/foldHelp"
+unsafeSqlCase _ (ECompositeKey _) = unexpectedCompositeKeyError "unsafeSqlCase"
 
 
 -- | (Internal) Create a custom binary operator.  You /should/
@@ -525,7 +607,7 @@ unsafeSqlBinOp op (ERaw p1 f1) (ERaw p2 f2) = ERaw Parens f
                  (b2, vals2) = f2 info
              in ( parensM p1 b1 <> op <> parensM p2 b2
                 , vals1 <> vals2 )
-unsafeSqlBinOp _ _ _ = unexpectedCompositeKeyError
+unsafeSqlBinOp _ _ _ = unexpectedCompositeKeyError "unsafeSqlBinOp"
 {-# INLINE unsafeSqlBinOp #-}
 
 
@@ -537,23 +619,46 @@ unsafeSqlBinOp _ _ _ = unexpectedCompositeKeyError
 --
 -- @
 -- (==.) :: SqlExpr (Value a) -> SqlExpr (Value a) -> SqlExpr (Value Bool)
--- (==.) = unsafeSqlBinOpList " = " " AND "
+-- (==.) = unsafeSqlBinOpComposite " = " " AND "
 -- @
-unsafeSqlBinOpList :: TLB.Builder -> TLB.Builder -> SqlExpr (Value a) -> SqlExpr (Value b) -> SqlExpr (Value c)
-unsafeSqlBinOpList op _ a@(ERaw _ _)    b@(ERaw _ _)  = unsafeSqlBinOp op a b
-unsafeSqlBinOpList op sep (ERawList f1) (ERawList f2) = ERaw Never f
+--
+-- Persistent has a hack for implementing composite keys (see
+-- 'ECompositeKey' doc for more details), so we're forced to use
+-- a hack here as well.  We deconstruct 'ERaw' values based on
+-- two rules:
+--
+--   - If it is a single placeholder, then it's assumed to be
+--   coming from a 'PersistList' and thus its components are
+--   separated so that they may be applied to a composite key.
+--
+--   - If it is not a single placeholder, then it's assumed to be
+--   a foreign (composite or not) key, so we enforce that it has
+--   no placeholders and split it on the commas.
+unsafeSqlBinOpComposite :: TLB.Builder -> TLB.Builder -> SqlExpr (Value a) -> SqlExpr (Value b) -> SqlExpr (Value c)
+unsafeSqlBinOpComposite op _ a@(ERaw _ _) b@(ERaw _ _) = unsafeSqlBinOp op a b
+unsafeSqlBinOpComposite op sep a b = ERaw Parens $ compose (listify a) (listify b)
   where
-    f info = let (b1, vals1) = f1 info
-                 (b2, vals2) = f2 info
-             in ( intersperseB sep . map (\(a,b) -> a <> op <> b) $
-                  zip b1 b2
-                , vals1 <> vals2 )
-unsafeSqlBinOpList op sep a b = unsafeSqlBinOpList op sep (promote a) (promote b)
-  where
-    promote :: SqlExpr (Value d) -> SqlExpr (Value d)
-    promote (ERaw p f)     = ERawList $ first (return . parensM p) . f
-    promote x@(ERawList _) = x
-{-# INLINE unsafeSqlBinOpList #-}
+    listify :: SqlExpr (Value x) -> IdentInfo -> ([TLB.Builder], [PersistValue])
+    listify (ECompositeKey f) = flip (,) [] . f
+    listify (ERaw _ f)        = deconstruct . f
+
+    deconstruct :: (TLB.Builder, [PersistValue]) -> ([TLB.Builder], [PersistValue])
+    deconstruct ("?", [PersistList vals]) = (replicate (length vals) "?", vals)
+    deconstruct (b, []) = (TLB.fromLazyText <$> TL.splitOn "," (TLB.toLazyText b), [])
+    deconstruct x = err $ "cannot deconstruct " ++ show x ++ "."
+
+    compose f1 f2 info
+      | not (null v1 || null v2) = err' "one side needs to have null placeholders"
+      | length b1 /= length b2   = err' "mismatching lengths"
+      | otherwise                = (bc, vc)
+      where
+        (b1, v1) = f1 info
+        (b2, v2) = f2 info
+        bc = intersperseB sep [x <> op <> y | (x, y) <- zip b1 b2]
+        vc = v1 <> v2
+        err' = err . (++ (", " ++ show ((b1, v1), (b2, v2))))
+
+    err = error . ("unsafeSqlBinOpComposite: " ++)
 
 -- | (Internal) A raw SQL value.  The same warning from
 -- 'unsafeSqlBinOp' applies to this function as well.
@@ -624,8 +729,8 @@ instance ( UnsafeSqlFunctionArgument a
 -- 'SqlExpr (Value b)'.  You should /not/ use this function
 -- unless you know what you're doing!
 veryUnsafeCoerceSqlExprValue :: SqlExpr (Value a) -> SqlExpr (Value b)
-veryUnsafeCoerceSqlExprValue (ERaw p f) = ERaw p f
-veryUnsafeCoerceSqlExprValue (ERawList f) = ERawList f
+veryUnsafeCoerceSqlExprValue (ERaw p f)        = ERaw p f
+veryUnsafeCoerceSqlExprValue (ECompositeKey f) = ECompositeKey f
 
 
 -- | (Internal) Coerce a value's type from 'SqlExpr (ValueList
@@ -710,12 +815,12 @@ selectSource query = do
 -- allow type information to flow both from @a@ to @r@ and
 -- vice-versa.  This means that you'll almost never have to give
 -- any type signatures for @esqueleto@ queries.  For example, the
--- query @select $ from $ \\p -> return p@ alone is ambiguous, but
+-- query @'select' $ from $ \\p -> return p@ alone is ambiguous, but
 -- in the context of
 --
 -- @
--- do ps <- select $
---          from $ \\p ->
+-- do ps <- 'select' $
+--          'from' $ \\p ->
 --          return p
 --    liftIO $ mapM_ (putStrLn . personName . entityVal) ps
 -- @
@@ -740,11 +845,8 @@ selectDistinctSource
      , MonadResource m )
   => SqlQuery a
   -> C.Source (SqlPersistT m) r
-selectDistinctSource query = do
-    src <- lift $ do
-        res <- rawSelectSource SELECT_DISTINCT query
-        fmap snd $ allocateAcquire res
-    src
+selectDistinctSource = selectSource . distinct
+{-# DEPRECATED selectDistinctSource "Since 2.2.4: use 'selectSource' and 'distinct'." #-}
 
 
 -- | Execute an @esqueleto@ @SELECT DISTINCT@ query inside
@@ -752,10 +854,8 @@ selectDistinctSource query = do
 selectDistinct :: ( SqlSelect a r
                   , MonadIO m )
                => SqlQuery a -> SqlPersistT m [r]
-selectDistinct query = do
-    res <- rawSelectSource SELECT_DISTINCT query
-    conn <- R.ask
-    liftIO $ with res $ flip R.runReaderT conn . runSource
+selectDistinct = select . distinct
+{-# DEPRECATED selectDistinct "Since 2.2.4: use 'select' and 'distinct'." #-}
 
 
 -- | (Internal) Run a 'C.Source' of rows.
@@ -789,9 +889,9 @@ rawEsqueleto mode query = do
 -- Example of usage:
 --
 -- @
--- delete $
--- from $ \\appointment ->
--- where_ (appointment ^. AppointmentDate <. val now)
+-- 'delete' $
+-- 'from' $ \\appointment ->
+-- 'where_' (appointment '^.' AppointmentDate '<.' 'val' now)
 -- @
 --
 -- Unlike 'select', there is a useful way of using 'delete' that
@@ -799,8 +899,8 @@ rawEsqueleto mode query = do
 -- (i.e., no 'where_' clause), you'll have to use a type signature:
 --
 -- @
--- delete $
--- from $ \\(appointment :: SqlExpr (Entity Appointment)) ->
+-- 'delete' $
+-- 'from' $ \\(appointment :: 'SqlExpr' ('Entity' Appointment)) ->
 -- return ()
 -- @
 delete :: ( MonadIO m )
@@ -824,9 +924,9 @@ deleteCount = rawEsqueleto DELETE
 -- Example of usage:
 --
 -- @
--- update $ \p -> do
--- set p [ PersonAge =. just (val thisYear) -. p ^. PersonBorn ]
--- where_ $ isNothing (p ^. PersonAge)
+-- 'update' $ \p -> do
+-- 'set' p [ PersonAge '=.' 'just' ('val' thisYear) -. p '^.' PersonBorn ]
+-- 'where_' $ isNothing (p '^.' PersonAge)
 -- @
 update :: ( MonadIO m
           , SqlEntity val )
@@ -864,13 +964,15 @@ toRawSql mode (conn, firstIdentState) query =
         flip S.runState firstIdentState $
         W.runWriterT $
         unQ query
-      SideData fromClauses
+      SideData distinctClause
+               fromClauses
                setClauses
                whereClauses
                groupByClause
                havingClause
                orderByClauses
-               limitClause = sd
+               limitClause
+               lockingClause = sd
       -- Pass the finalIdentState (containing all identifiers
       -- that were used) to the subsequent calls.  This ensures
       -- that no name clashes will occur on subqueries that may
@@ -878,7 +980,7 @@ toRawSql mode (conn, firstIdentState) query =
       info = (conn, finalIdentState)
   in mconcat
       [ makeInsertInto info mode ret
-      , makeSelect     info mode ret
+      , makeSelect     info mode distinctClause ret
       , makeFrom       info mode fromClauses
       , makeSet        info setClauses
       , makeWhere      info whereClauses
@@ -886,17 +988,16 @@ toRawSql mode (conn, firstIdentState) query =
       , makeHaving     info havingClause
       , makeOrderBy    info orderByClauses
       , makeLimit      info limitClause orderByClauses
+      , makeLocking         lockingClause
       ]
 
 
 -- | (Internal) Mode of query being converted by 'toRawSql'.
 data Mode =
     SELECT
-  | SELECT_DISTINCT
   | DELETE
   | UPDATE
-  | INSERT_INTO Mode
-    -- ^ 'Mode' should be either 'SELECT' or 'SELECT_DISTINCT'.
+  | INSERT_INTO
 
 
 uncommas :: [TLB.Builder] -> TLB.Builder
@@ -910,21 +1011,27 @@ uncommas' = (uncommas *** mconcat) . unzip
 
 
 makeInsertInto :: SqlSelect a r => IdentInfo -> Mode -> a -> (TLB.Builder, [PersistValue])
-makeInsertInto info (INSERT_INTO _) ret = sqlInsertInto info ret
-makeInsertInto _    _               _   = mempty
+makeInsertInto info INSERT_INTO ret = sqlInsertInto info ret
+makeInsertInto _    _           _   = mempty
 
 
-makeSelect :: SqlSelect a r => IdentInfo -> Mode -> a -> (TLB.Builder, [PersistValue])
-makeSelect info mode_ ret = process mode_
+makeSelect :: SqlSelect a r => IdentInfo -> Mode -> DistinctClause -> a -> (TLB.Builder, [PersistValue])
+makeSelect info mode_ distinctClause ret = process mode_
   where
     process mode =
       case mode of
-        SELECT            -> withCols "SELECT "
-        SELECT_DISTINCT   -> withCols "SELECT DISTINCT "
-        DELETE            -> plain "DELETE "
-        UPDATE            -> plain "UPDATE "
-        INSERT_INTO mode' -> process mode'
-    withCols v = first (v <>) (sqlSelectCols info ret)
+        SELECT      -> withCols selectKind
+        DELETE      -> plain "DELETE "
+        UPDATE      -> plain "UPDATE "
+        INSERT_INTO -> process SELECT
+    selectKind =
+      case distinctClause of
+        DistinctAll      -> ("SELECT ", [])
+        DistinctStandard -> ("SELECT DISTINCT ", [])
+        DistinctOn exprs -> first (("SELECT DISTINCT ON (" <>) . (<> ") ")) $
+                            uncommas' (processExpr <$> exprs)
+      where processExpr (EDistinctOn f) = materializeExpr info f
+    withCols v = v <> (sqlSelectCols info ret)
     plain    v = (v, [])
 
 
@@ -934,19 +1041,20 @@ makeFrom info mode fs = ret
   where
     ret = case collectOnClauses fs of
             Left expr -> throw $ mkExc expr
-            Right fs' -> keyword $ uncommas' (map (mk Never mempty) fs')
+            Right fs' -> keyword $ uncommas' (map (mk Never) fs')
     keyword = case mode of
                 UPDATE -> id
                 _      -> first ("\nFROM " <>)
 
-    mk _     onClause (FromStart i def) = base i def <> onClause
-    mk paren onClause (FromJoin lhs kind rhs monClause) =
+    mk _     (FromStart i def) = base i def
+    mk paren (FromJoin lhs kind rhs monClause) =
       first (parensM paren) $
-      mconcat [ mk Parens onClause lhs
+      mconcat [ mk Never lhs
               , (fromKind kind, mempty)
-              , mk Never (maybe mempty makeOnClause monClause) rhs
+              , mk Parens rhs
+              , maybe mempty makeOnClause monClause
               ]
-    mk _ _ (OnClause _) = error "Esqueleto/Sql/makeFrom: never here (is collectOnClauses working?)"
+    mk _ (OnClause _) = error "Esqueleto/Sql/makeFrom: never here (is collectOnClauses working?)"
 
     base ident@(I identText) def =
       let db@(DBName dbText) = entityDB def
@@ -961,30 +1069,30 @@ makeFrom info mode fs = ret
     fromKind RightOuterJoinKind = " RIGHT OUTER JOIN "
     fromKind FullOuterJoinKind  = " FULL OUTER JOIN "
 
-    makeOnClause (ERaw _ f)   = first (" ON " <>) (f info)
-    makeOnClause (ERawList _) = unexpectedCompositeKeyError
+    makeOnClause (ERaw _ f)        = first (" ON " <>) (f info)
+    makeOnClause (ECompositeKey _) = unexpectedCompositeKeyError "makeFrom/makeOnClause"
 
     mkExc :: SqlExpr (Value Bool) -> OnClauseWithoutMatchingJoinException
     mkExc (ERaw _ f) =
       OnClauseWithoutMatchingJoinException $
       TL.unpack $ TLB.toLazyText $ fst (f info)
-    mkExc (ERawList _) = unexpectedCompositeKeyError
+    mkExc (ECompositeKey _) = unexpectedCompositeKeyError "makeFrom/mkExc"
 
-unexpectedCompositeKeyError :: a
-unexpectedCompositeKeyError = error "non-id/composite keys not expected here"
+unexpectedCompositeKeyError :: String -> a
+unexpectedCompositeKeyError w = error $ w ++ ": non-id/composite keys not expected here"
 
 
 makeSet :: IdentInfo -> [SetClause] -> (TLB.Builder, [PersistValue])
 makeSet _    [] = mempty
 makeSet info os = first ("\nSET " <>) . uncommas' $ concatMap mk os
   where
-    mk (SetClause (ERaw _ f))   = [f info]
-    mk (SetClause (ERawList _)) = error "esqueleto/makeSet: FIXME"
+    mk (SetClause (ERaw _ f))        = [f info]
+    mk (SetClause (ECompositeKey _)) = unexpectedCompositeKeyError "makeSet" -- FIXME
 
 makeWhere :: IdentInfo -> WhereClause -> (TLB.Builder, [PersistValue])
-makeWhere _    NoWhere              = mempty
-makeWhere info (Where (ERaw _ f))   = first ("\nWHERE " <>) (f info)
-makeWhere _    (Where (ERawList _)) = unexpectedCompositeKeyError
+makeWhere _    NoWhere                   = mempty
+makeWhere info (Where (ERaw _ f))        = first ("\nWHERE " <>) (f info)
+makeWhere _    (Where (ECompositeKey _)) = unexpectedCompositeKeyError "makeWhere"
 
 
 makeGroupBy :: IdentInfo -> GroupByClause -> (TLB.Builder, [PersistValue])
@@ -994,19 +1102,20 @@ makeGroupBy info (GroupBy fields) = first ("\nGROUP BY " <>) build
     build = uncommas' $ map (\(SomeValue (ERaw _ f)) -> f info) fields
 
 makeHaving :: IdentInfo -> WhereClause -> (TLB.Builder, [PersistValue])
-makeHaving _    NoWhere               = mempty
-makeHaving info (Where (ERaw _ f))    = first ("\nHAVING " <>) (f info)
-makeHaving _    (Where (ERawList _ )) = unexpectedCompositeKeyError
+makeHaving _    NoWhere                    = mempty
+makeHaving info (Where (ERaw _ f))         = first ("\nHAVING " <>) (f info)
+makeHaving _    (Where (ECompositeKey _ )) = unexpectedCompositeKeyError "makeHaving"
 
 makeOrderBy :: IdentInfo -> [OrderByClause] -> (TLB.Builder, [PersistValue])
 makeOrderBy _    [] = mempty
 makeOrderBy info os = first ("\nORDER BY " <>) . uncommas' $ concatMap mk os
   where
     mk :: OrderByClause -> [(TLB.Builder, [PersistValue])]
-    mk (EOrderBy t (ERaw p f))   = [first ((<> orderByType t) . parensM p) (f info)]
-    mk (EOrderBy t (ERawList f)) = let fs = fst $ f info
-                                       vals = repeat []
-                                   in zip (map (<> orderByType t) fs) vals
+    mk (EOrderBy t (ERaw p f)) = [first ((<> orderByType t) . parensM p) (f info)]
+    mk (EOrderBy t (ECompositeKey f)) =
+      let fs = f info
+          vals = repeat []
+      in zip (map (<> orderByType t) fs) vals
     mk EOrderRandom = [first ((<> "RANDOM()")) mempty]
     orderByType ASC  = " ASC"
     orderByType DESC = " DESC"
@@ -1020,6 +1129,15 @@ makeLimit (conn, _) (Limit ml mo) orderByClauses =
   in (TLB.fromText limitRaw, mempty)
 
 
+makeLocking :: LockingClause -> (TLB.Builder, [PersistValue])
+makeLocking = flip (,) [] . maybe mempty toTLB . getLast
+  where
+    toTLB ForUpdate       = "\nFOR UPDATE"
+    toTLB ForShare        = "\nFOR SHARE"
+    toTLB LockInShareMode = "\nLOCK IN SHARE MODE"
+
+
+
 parens :: TLB.Builder -> TLB.Builder
 parens b = "(" <> (b <> ")")
 
@@ -1113,15 +1231,20 @@ instance PersistEntity a => SqlSelect (SqlExpr (Maybe (Entity a))) (Maybe (Entit
 -- | You may return any single value (i.e. a single column) from
 -- a 'select' query.
 instance PersistField a => SqlSelect (SqlExpr (Value a)) (Value a) where
-  sqlSelectCols info (ERaw p f) = let (b, vals) = f info
-                                  in (parensM p b, vals)
-  sqlSelectCols info (ERawList f) = let (b, vals) = f info
-                                    in case b of
-                                         [b'] -> (parensM Parens b', vals)
-                                         _    -> unexpectedCompositeKeyError
+  sqlSelectCols = materializeExpr
   sqlSelectColCount = const 1
   sqlSelectProcessRow [pv] = Value <$> fromPersistValue pv
-  sqlSelectProcessRow _    = Left "SqlSelect (Value a): wrong number of columns."
+  sqlSelectProcessRow pvs  = Value <$> fromPersistValue (PersistList pvs)
+
+
+-- | Materialize a @SqlExpr (Value a)@.
+materializeExpr :: IdentInfo -> SqlExpr (Value a) -> (TLB.Builder, [PersistValue])
+materializeExpr info (ERaw p f) =
+  let (b, vals) = f info
+  in (parensM p b, vals)
+materializeExpr info (ECompositeKey f) =
+  let bs = f info
+  in (uncommas $ map (parensM Parens) bs, [])
 
 
 -- | You may return tuples (up to 16-tuples) and tuples of tuples
@@ -1618,16 +1741,11 @@ to16 ((a,b),(c,d),(e,f),(g,h),(i,j),(k,l),(m,n),(o,p)) = (a,b,c,d,e,f,g,h,i,j,k,
 -- | Insert a 'PersistField' for every selected value.
 insertSelect :: (MonadIO m, PersistEntity a) =>
   SqlQuery (SqlExpr (Insertion a)) -> SqlPersistT m ()
-insertSelect = insertGeneralSelect SELECT
+insertSelect = liftM (const ()) . rawEsqueleto INSERT_INTO . fmap EInsertFinal
 
 
 -- | Insert a 'PersistField' for every unique selected value.
 insertSelectDistinct :: (MonadIO m, PersistEntity a) =>
   SqlQuery (SqlExpr (Insertion a)) -> SqlPersistT m ()
-insertSelectDistinct = insertGeneralSelect SELECT_DISTINCT
-
-
-insertGeneralSelect :: (MonadIO m, PersistEntity a) =>
-  Mode -> SqlQuery (SqlExpr (Insertion a)) -> SqlPersistT m ()
-insertGeneralSelect mode =
-  liftM (const ()) . rawEsqueleto (INSERT_INTO mode) . fmap EInsertFinal
+insertSelectDistinct = insertSelect . distinct
+{-# DEPRECATED insertSelectDistinct "Since 2.2.4: use 'insertSelect' and 'distinct'." #-}
