@@ -1,3 +1,4 @@
+{-# OPTIONS_GHC -fno-warn-unused-binds  #-}
 {-# LANGUAGE ConstraintKinds
            , EmptyDataDecls
            , FlexibleContexts
@@ -18,12 +19,17 @@
 module Main (main) where
 
 import Control.Applicative ((<$>))
+import Control.Arrow ((&&&))
 import Control.Exception (IOException)
-import Control.Monad (replicateM, replicateM_)
+import Control.Monad (forM_, replicateM, replicateM_, void)
 import Control.Monad.IO.Class (MonadIO(liftIO))
 import Control.Monad.Logger (MonadLogger(..), runStderrLoggingT, runNoLoggingT)
 import Control.Monad.Trans.Control (MonadBaseControl(..))
 import Control.Monad.Trans.Reader (ReaderT)
+import Data.Char (toLower, toUpper)
+import Data.List (sortBy)
+import Data.Monoid ((<>))
+import Data.Ord (comparing)
 import Database.Esqueleto
 #if   defined (WITH_POSTGRESQL)
 import Database.Persist.Postgresql (withPostgresqlConn)
@@ -44,13 +50,21 @@ import Database.Persist.TH
 import Test.Hspec
 
 import qualified Control.Monad.Trans.Resource as R
-import qualified Data.Set as S
 import qualified Data.List as L
-import Data.Char (toLower, toUpper)
+import qualified Data.Set as S
+import qualified Data.Text.Lazy.Builder as TLB
+import qualified Database.Esqueleto.PostgreSQL as EP
+import qualified Database.Esqueleto.Internal.Sql as EI
 
 
 -- Test schema
 share [mkPersist sqlSettings, mkMigrate "migrateAll"] [persistUpperCase|
+  Foo
+    name Int
+    Primary name
+  Bar
+    quux FooId
+
   Person
     name String
     age Int Maybe
@@ -65,6 +79,10 @@ share [mkPersist sqlSettings, mkMigrate "migrateAll"] [persistUpperCase|
     follower PersonId
     followed PersonId
     deriving Eq Show
+
+  CcList
+    names [String]
+
   Frontcover
     number Int
     title String
@@ -100,6 +118,9 @@ share [mkPersist sqlSettings, mkMigrate "migrateAll"] [persistUpperCase|
     name String
     Foreign Point fkpoint centerX centerY
     deriving Eq Show
+  Numbers
+    int    Int
+    double Double
 |]
 
 -- | this could be achieved with S.fromList, but not all lists
@@ -285,14 +306,11 @@ main = do
             thePk `shouldBe` tagPk
 
       it "works when returning a composite primary key from a query" $
-        pendingWith "Need to refactor 'Value a's SqlQuery instance"
-        {-
         run $ do
           let p = Point 10 20 ""
           thePk <- insert p
           [Value ppk] <- select $ from $ \p' -> return (p'^.PointId)
           liftIO $ ppk `shouldBe` thePk
-        -}
 
 
     describe "select/JOIN" $ do
@@ -439,6 +457,19 @@ main = do
             retArt `shouldBe` article
             retTag `shouldBe` tag
 
+      it "respects the associativity of joins" $
+        run $ do
+            void $ insert p1
+            ps <- select . from $
+                      \((p :: SqlExpr (Entity Person))
+                       `LeftOuterJoin`
+                        ((_q :: SqlExpr (Entity Person))
+                         `InnerJoin` (_r :: SqlExpr (Entity Person)))) -> do
+                on (val False) -- Inner join is empty
+                on (val True)
+                return p
+            liftIO $ (entityVal <$> ps) `shouldBe` [p1]
+
     describe "select/where_" $ do
       it "works for a simple example with (==.)" $
         run $ do
@@ -540,18 +571,18 @@ main = do
           p2e@(Entity _ bob) <- insert' $ Person "bob" (Just 36) Nothing   1
 
           -- lower(name) == 'john'
-          ret <- select $
-                 from $ \p-> do
-                 where_ (lower_ (p ^. PersonName) ==. val (map toLower $ personName p1))
-                 return p
-          liftIO $ ret `shouldBe` [ p1e ]
+          ret1 <- select $
+                  from $ \p-> do
+                  where_ (lower_ (p ^. PersonName) ==. val (map toLower $ personName p1))
+                  return p
+          liftIO $ ret1 `shouldBe` [ p1e ]
 
           -- name == lower('BOB')
-          ret <- select $
-                 from $ \p-> do
-                 where_ (p ^. PersonName ==. lower_ (val $ map toUpper $ personName bob))
-                 return p
-          liftIO $ ret `shouldBe` [ p2e ]
+          ret2 <- select $
+                  from $ \p-> do
+                  where_ (p ^. PersonName ==. lower_ (val $ map toUpper $ personName bob))
+                  return p
+          liftIO $ ret2 `shouldBe` [ p2e ]
 
       it "works with random_" $
         run $ do
@@ -682,7 +713,7 @@ main = do
                  return p
           liftIO $ ret `shouldBe` [ p1e, p3e, p2e ]
 
-      it "works with two ASC fields" $
+      it "works with two ASC fields (one call)" $
         run $ do
           p1e <- insert' p1
           p2e <- insert' p2
@@ -700,7 +731,7 @@ main = do
           liftIO $ ret `shouldBe` [ p2e, p4e, p3e, p1e ]
 #endif
 
-      it "works with one ASC and one DESC field" $
+      it "works with one ASC and one DESC field (two calls)" $
         run $ do
           p1e <- insert' p1
           p2e <- insert' p2
@@ -708,7 +739,8 @@ main = do
           p4e <- insert' p4
           ret <- select $
                  from $ \p -> do
-                 orderBy [desc (p ^. PersonAge), asc (p ^. PersonName)]
+                 orderBy [desc (p ^. PersonAge)]
+                 orderBy [asc (p ^. PersonName)]
                  return p
 #ifdef WITH_POSTGRESQL
           liftIO $ ret `shouldBe` [ p2e, p1e, p4e, p3e ]
@@ -759,18 +791,75 @@ main = do
           liftIO $ map entityVal eps `shouldBe` reverse ps
 
 
-    describe "selectDistinct" $
-      it "works on a simple example" $
+    describe "SELECT DISTINCT" $ do
+      let selDistTest
+            :: (   forall m. RunDbMonad m
+                => SqlQuery (SqlExpr (Value String))
+                -> SqlPersistT (R.ResourceT m) [Value String])
+            -> IO ()
+          selDistTest q =
+            run $ do
+              p1k <- insert p1
+              let (t1, t2, t3) = ("a", "b", "c")
+              mapM_ (insert . flip BlogPost p1k) [t1, t3, t2, t2, t1]
+              ret <- q $
+                     from $ \b -> do
+                     let title = b ^. BlogPostTitle
+                     orderBy [asc title]
+                     return title
+              liftIO $ ret `shouldBe` [ Value t1, Value t2, Value t3 ]
+      it "works on a simple example (selectDistinct)" $
+        selDistTest selectDistinct
+
+      it "works on a simple example (select . distinct)" $
+        selDistTest (select . distinct)
+
+      it "works on a simple example (distinct (return ()))" $
+        selDistTest (\act -> select $ distinct (return ()) >> act)
+
+#if defined(WITH_POSTGRESQL)
+    describe "SELECT DISTINCT ON" $ do
+      it "works on a simple example" $ do
         run $ do
-          p1k <- insert p1
-          let (t1, t2, t3) = ("a", "b", "c")
-          mapM_ (insert . flip BlogPost p1k) [t1, t3, t2, t2, t1]
-          ret <- selectDistinct $
-                 from $ \b -> do
-                 let title = b ^. BlogPostTitle
-                 orderBy [asc title]
-                 return title
-          liftIO $ ret `shouldBe` [ Value t1, Value t2, Value t3 ]
+          [p1k, p2k, _] <- mapM insert [p1, p2, p3]
+          [_, bpB, bpC] <- mapM insert'
+            [ BlogPost "A" p1k
+            , BlogPost "B" p1k
+            , BlogPost "C" p2k ]
+          ret <- select $
+                 from $ \bp ->
+                 distinctOn [don (bp ^. BlogPostAuthorId)] $ do
+                 orderBy [asc (bp ^. BlogPostAuthorId), desc (bp ^. BlogPostTitle)]
+                 return bp
+          liftIO $ ret `shouldBe` sortBy (comparing (blogPostAuthorId . entityVal)) [bpB, bpC]
+
+      let slightlyLessSimpleTest q =
+            run $ do
+              [p1k, p2k, _] <- mapM insert [p1, p2, p3]
+              [bpA, bpB, bpC] <- mapM insert'
+                [ BlogPost "A" p1k
+                , BlogPost "B" p1k
+                , BlogPost "C" p2k ]
+              ret <- select $
+                     from $ \bp ->
+                     q bp $ return bp
+              let cmp = (blogPostAuthorId &&& blogPostTitle) . entityVal
+              liftIO $ ret `shouldBe` sortBy (comparing cmp) [bpA, bpB, bpC]
+      it "works on a slightly less simple example (two distinctOn calls, orderBy)" $
+        slightlyLessSimpleTest $ \bp act ->
+          distinctOn [don (bp ^. BlogPostAuthorId)] $
+          distinctOn [don (bp ^. BlogPostTitle)] $ do
+            orderBy [asc (bp ^. BlogPostAuthorId), asc (bp ^. BlogPostTitle)]
+            act
+      it "works on a slightly less simple example (one distinctOn call, orderBy)" $ do
+        slightlyLessSimpleTest $ \bp act ->
+          distinctOn [don (bp ^. BlogPostAuthorId), don (bp ^. BlogPostTitle)] $ do
+            orderBy [asc (bp ^. BlogPostAuthorId), asc (bp ^. BlogPostTitle)]
+            act
+      it "works on a slightly less simple example (distinctOnOrderBy)" $ do
+        slightlyLessSimpleTest $ \bp ->
+          distinctOnOrderBy [asc (bp ^. BlogPostAuthorId), asc (bp ^. BlogPostTitle)]
+#endif
 
     describe "coalesce/coalesceDefault" $ do
       it "works on a simple example" $
@@ -823,10 +912,11 @@ main = do
 #if defined(WITH_POSTGRESQL) || defined(WITH_MYSQL)
       it "works on PostgreSQL and MySQL with <2 arguments" $
         run $ do
-          _ :: [Value (Maybe Int)] <- select $
-               from $ \p -> do
-               return (coalesce [p ^. PersonAge])
-          return True
+          _ :: [Value (Maybe Int)] <-
+            select $
+            from $ \p -> do
+            return (coalesce [p ^. PersonAge])
+          return ()
 #else
       it "throws an exception on SQLite with <2 arguments" $
         run (select $
@@ -839,7 +929,7 @@ main = do
 #endif
 #endif
 
-    describe "text functions" $
+    describe "text functions" $ do
       it "like, (%) and (++.) work on a simple example" $
          run $ do
            [p1e, p2e, p3e, p4e] <- mapM insert' [p1, p2, p3, p4]
@@ -854,6 +944,21 @@ main = do
            nameContains "i"  [p4e, p3e]
            nameContains "iv" [p4e]
 
+#if defined(WITH_POSTGRESQL)
+      it "ilike, (%) and (++.) work on a simple example on PostgreSQL" $
+         run $ do
+           [p1e, _, p3e, _, p5e] <- mapM insert' [p1, p2, p3, p4, p5]
+           let nameContains t expected = do
+                 ret <- select $
+                        from $ \p -> do
+                        where_ (p ^. PersonName `ilike` (%) ++. val t ++. (%))
+                        orderBy [asc (p ^. PersonName)]
+                        return p
+                 liftIO $ ret `shouldBe` expected
+           nameContains "mi" [p3e, p5e]
+           nameContains "JOHN" [p1e]
+#endif
+
     describe "delete" $
       it "works on a simple example" $
         run $ do
@@ -1062,6 +1167,7 @@ main = do
                  where_ $ exists $
                           from $ \bp -> do
                           where_ (bp ^. BlogPostAuthorId ==. p ^. PersonId)
+                 orderBy [asc (p ^. PersonName)]
                  return p
           liftIO $ ret `shouldBe` [ Entity p1k p1
                                   , Entity p3k p3 ]
@@ -1081,6 +1187,16 @@ main = do
                  return p
           liftIO $ ret `shouldBe` [ Entity p2k p2 ]
 
+
+    describe "list fields" $ do
+      -- <https://github.com/prowdsponsor/esqueleto/issues/100>
+      it "can update list fields" $
+        run $ do
+          cclist <- insert $ CcList []
+          update $ \p -> do
+            set p [ CcListNames =. val ["fred"]]
+            where_ (p ^. CcListId ==. val cclist)
+
     describe "inserts by select" $ do
       it "IN works for insertSelect" $
         run $ do
@@ -1092,8 +1208,8 @@ main = do
           ret <- select $ from (\(_::(SqlExpr (Entity BlogPost))) -> return countRows)
           liftIO $ ret `shouldBe` [Value (3::Int)]
 
-    describe "rand works" $ do
-      it "returns result in random order" $
+    describe "Math-related functions" $ do
+      it "rand returns result in random order" $
         run $ do
           replicateM_ 20 $ do
             _ <- insert p1
@@ -1113,6 +1229,19 @@ main = do
 
           liftIO $ (ret1 == ret2) `shouldBe` False
 
+      it "castNum works for multiplying Int and Double" $
+        run $ do
+          mapM_ insert [Numbers 2 3.4, Numbers 7 1.1]
+          ret <-
+            select $
+            from $ \n -> do
+            let r = castNum (n ^. NumbersInt) *. n ^. NumbersDouble
+            orderBy [asc r]
+            return r
+          liftIO $ length ret `shouldBe` 2
+          let [Value a, Value b] = ret
+          liftIO $ max (abs (a - 6.8)) (abs (b - 7.7)) `shouldSatisfy` (< 0.01)
+
     describe "case" $ do
       it "Works for a simple value based when - False" $
         run $ do
@@ -1159,6 +1288,97 @@ main = do
 
           liftIO $ ret `shouldBe` [ Value (3) ]
 
+    describe "locking" $ do
+      -- The locking clause is the last one, so try to use many
+      -- others to test if it's at the right position.  We don't
+      -- care about the text of the rest, nor with the RDBMS'
+      -- reaction to the clause.
+      let sanityCheck kind syntax = do
+            let complexQuery =
+                  from $ \(p1 `InnerJoin` p2) -> do
+                  on (p1 ^. PersonName ==. p2 ^. PersonName)
+                  where_ (p1 ^. PersonFavNum >. val 2)
+                  orderBy [desc (p2 ^. PersonAge)]
+                  limit 3
+                  offset 9
+                  groupBy (p1 ^. PersonId)
+                  having (countRows <. val (0 :: Int))
+                  return (p1, p2)
+                queryWithClause1 = do
+                  r <- complexQuery
+                  locking kind
+                  return r
+                queryWithClause2 = do
+                  locking ForUpdate
+                  r <- complexQuery
+                  locking ForShare
+                  locking kind
+                  return r
+                queryWithClause3 = do
+                  locking kind
+                  complexQuery
+                toText conn q =
+                  let (tlb, _) = EI.toRawSql EI.SELECT (conn, EI.initialIdentState) q
+                  in TLB.toLazyText tlb
+            [complex, with1, with2, with3] <-
+              runNoLoggingT $ withConn $ \conn -> return $
+                map (toText conn) [complexQuery, queryWithClause1, queryWithClause2, queryWithClause3]
+            let expected = complex <> "\n" <> syntax
+            (with1, with2, with3) `shouldBe` (expected, expected, expected)
+
+      it "looks sane for ForUpdate"       $ sanityCheck ForUpdate       "FOR UPDATE"
+      it "looks sane for ForShare"        $ sanityCheck ForShare        "FOR SHARE"
+      it "looks sane for LockInShareMode" $ sanityCheck LockInShareMode "LOCK IN SHARE MODE"
+
+    describe "counting rows" $ do
+      forM_ [ ("count (test A)",    count . (^. PersonAge),         4)
+            , ("count (test B)",    count . (^. PersonWeight),      5)
+            , ("countRows",         const countRows,                5)
+            , ("countDistinct",     countDistinct . (^. PersonAge), 2) ] $
+        \(title, countKind, expected) ->
+        it (title ++ " works as expected") $
+          run $ do
+            mapM_ insert
+              [ Person "" (Just 1) (Just 1) 1
+              , Person "" (Just 2) (Just 1) 1
+              , Person "" (Just 2) (Just 1) 1
+              , Person "" (Just 2) (Just 2) 1
+              , Person "" Nothing  (Just 3) 1]
+            [Value n] <- select $ from $ return . countKind
+            liftIO $ (n :: Int) `shouldBe` expected
+
+    describe "PostgreSQL module" $ do
+      it "should be tested on the PostgreSQL database" $
+#if !defined(WITH_POSTGRESQL)
+        pendingWith "test suite not running under PostgreSQL, skipping"
+#else
+        (return () :: IO ())
+
+      it "arrayAgg looks sane" $
+        run $ do
+          let people = [p1, p2, p3, p4, p5]
+          mapM_ insert people
+          [Value ret] <-
+            select $
+            from $ \p -> do
+            return (EP.arrayAgg (p ^. PersonName))
+          liftIO $ L.sort ret `shouldBe` L.sort (map personName people)
+
+      it "stringAgg looks sane" $
+        run $ do
+          let people = [p1, p2, p3, p4, p5]
+          mapM_ insert people
+          [Value ret] <-
+            select $
+            from $ \p -> do
+            return (EP.stringAgg (p ^. PersonName) (val " "))
+          liftIO $ L.sort (words ret) `shouldBe` L.sort (map personName people)
+
+      it "chr looks sane" $
+        run $ do
+          [Value (ret :: String)] <- select $ return (EP.chr (val 65))
+          liftIO $ ret `shouldBe` "A"
+#endif
 
 ----------------------------------------------------------------------
 
@@ -1183,9 +1403,20 @@ cleanDB
   :: (forall m. RunDbMonad m
   => SqlPersistT (R.ResourceT m) ())
 cleanDB = do
-  delete $ from $ \(blogpost :: SqlExpr (Entity BlogPost))-> return ()
-  delete $ from $ \(follow   :: SqlExpr (Entity Follow))  -> return ()
-  delete $ from $ \(person   :: SqlExpr (Entity Person))  -> return ()
+  delete $ from $ \(_ :: SqlExpr (Entity BlogPost))   -> return ()
+  delete $ from $ \(_ :: SqlExpr (Entity Follow))     -> return ()
+  delete $ from $ \(_ :: SqlExpr (Entity Person))     -> return ()
+
+  delete $ from $ \(_ :: SqlExpr (Entity ArticleTag)) -> return ()
+  delete $ from $ \(_ :: SqlExpr (Entity Article))    -> return ()
+  delete $ from $ \(_ :: SqlExpr (Entity Article2))   -> return ()
+  delete $ from $ \(_ :: SqlExpr (Entity Tag))        -> return ()
+  delete $ from $ \(_ :: SqlExpr (Entity Frontcover)) -> return ()
+
+  delete $ from $ \(_ :: SqlExpr (Entity Circle))     -> return ()
+  delete $ from $ \(_ :: SqlExpr (Entity Point))      -> return ()
+
+  delete $ from $ \(_ :: SqlExpr (Entity Numbers))    -> return ()
 #endif
 
 
@@ -1203,23 +1434,29 @@ verbose = True
 
 
 run_worker :: RunDbMonad m => SqlPersistT (R.ResourceT m) a -> m a
-run_worker act =
+run_worker act = withConn $ runSqlConn (migrateIt >> act)
+
+
+migrateIt :: RunDbMonad m => SqlPersistT (R.ResourceT m) ()
+migrateIt = do
+  void $ runMigrationSilent migrateAll
+#if defined (WITH_POSTGRESQL) || defined (WITH_MYSQL)
+  cleanDB
+#endif
+
+
+withConn :: RunDbMonad m => (SqlBackend -> R.ResourceT m a) -> m a
+withConn =
   R.runResourceT .
 #if defined(WITH_POSTGRESQL)
-  withPostgresqlConn "host=localhost port=5432 user=test dbname=test" .
+  withPostgresqlConn "host=localhost port=5432 user=test dbname=test"
 #elif defined (WITH_MYSQL)
   withMySQLConn defaultConnectInfo
     { connectHost     = "localhost"
     , connectUser     = "test"
     , connectPassword = "test"
     , connectDatabase = "test"
-    } .
+    }
 #else
-  withSqliteConn ":memory:" .
-#endif
-  runSqlConn .
-#if defined (WITH_POSTGRESQL) || defined (WITH_MYSQL)
-  (runMigrationSilent migrateAll >>) $ (cleanDB >> act)
-#else
-  (runMigrationSilent migrateAll >>) $ act
+  withSqliteConn ":memory:"
 #endif
